<?php

namespace Framework {

	abstract class FilterArgument {
		public function __construct($name) {
			$name = ucfirst($name);

			$this->name = $name;

			if (isset($_GET[$name])) {
				$this->value = $_GET[$name];
			} else {
				$this->value = false;
			}
		}

		public function isPresent() {
			return $this->value !== false;
		}

		public function getName() {
			return $this->name;
		}

		abstract public function sqlCondition($columnName);

		public function sqlParameters() {
			return array(&$this->value);
		}

		public function sqlParamTypes() {
			return 's';
		}
		
		protected $name;
		protected $value;
	}


	class FilterEquals extends FilterArgument {
		private function isNull() {
			return $this->value === 'null';
		}
	
		public function sqlParamTypes() {
			if ($this->isNull()) {
				return '';
			}
			
			return parent::sqlParamTypes();
		}
		
		public function sqlParameters() {
			if ($this->isNull()) {
				return array();
			}
			
			return parent::sqlParameters();
		}
		
		public function sqlCondition($columnName) {
			if ($this->isNull()) {
				return ' AND ' . $columnName . ' IS NULL';
			}
			
			return ' AND ' . $columnName . ' = ?';
		}
	}


	class FilterContains extends FilterArgument {
		public function sqlCondition($columnName) {
			return ' AND ' . $columnName . ' LIKE ?';
		}

		public function sqlParameters() {
			if ($this->pattern == null) {
				$this->pattern = '%' . str_replace('%', '\\%', $this->value) . '%';
			}

			return array(&$this->pattern);
		}

		private $pattern;
	}

	class FilterOneOf extends FilterArgument {
		public function __construct($name) {
			parent::__construct($name);
			
			if ($this->value !== null) {
				$this->value = explode(',', $this->value);
			}
		}

		public function sqlCondition($columnName) {
			$condition = ' AND ' . $columnName 
				. ' IN ('
				. implode(', ', array_fill(0, count($this->value), '?'))
				. ')';

			return $condition;
		}

		public function sqlParameters() {
			$params = array();

			foreach ($this->value as $index => $value) {
				$params[] = &$this->value[$index];
			}

			return $params;
		}

		public function sqlParamTypes() {
			return str_repeat('s', count($this->value));
		}
	}

	class FilterRangeFrom extends FilterArgument {
		public function sqlCondition($columnName) {
			return ' AND ' . $columnName . ' >= ?';
		}
	}

	class FilterRangeTo extends FilterArgument {
		public function sqlCondition($columnName) {
			return ' AND ' . $columnName . ' <= ?';
		}
	}
	
	class FilterSearchString extends FilterArgument {
		public function __construct($paramName, $searchColumns) {
			parent::__construct($paramName);
			
			if ($this->value !== null) {
				$this->value = array_filter(str_getcsv($this->value, ' '), function ($value) { return $value != ''; });
			}
			
			$this->searchColumns = $searchColumns;
		}
		
		public function sqlCondition($columnNames) {
			$condition = '';
			$outerSeparator = '';
			
			$phraseCount = count($this->value);
			
			for ($i = 0; $i < $phraseCount; $i++) {
				$separator = '';
				$condition .= $outerSeparator;
				$condition .= '(';
				foreach ($this->searchColumns as $name) {
					if (isset($columnNames[$name])) {
						$condition .= $separator;
						$condition .= $columnNames[$name];
						$condition .= ' LIKE ?';
						
						$separator = ' OR ';
					}
				}
				
				$condition .= ')';
				$outerSeparator = ' AND ';
			}
			
			if ($condition != '') {
				$condition = ' AND (' . $condition . ')';
			}
			
			return $condition;
		}
		
		public function sqlParamTypes() {
			return str_repeat('s', count($this->searchColumns) * count($this->value));
		}
		
		public function sqlParameters() {
			if ($this->patterns === null) {
				$this->patterns = array();
				foreach ($this->value as $phrase) {
					$pattern = '%' . str_replace('%', '\\%', $phrase) . '%';
					$this->patterns[] = $pattern;
				}
			}
			
			$params = array();
			
			foreach ($this->patterns as $patternIndex => $pattern) {
				for ($i = 0; $i < count($this->searchColumns); $i++) {
					$params[] = &$this->patterns[$patternIndex];
				}
			}
			
			return $params;
		}
		
		private $searchColumns;
		private $pattern;
	}

	class Filter {
		const Equals       = 1;
		const Contains     = 2;
		const OneOf        = 3;
		const RangeFrom    = 4;
		const RangeTo      = 5;
		const SearchString = 6;
		
		private static $filterClassNames = array(
			Filter::Equals       => 'FilterEquals',
			Filter::Contains     => 'FilterContains',
			Filter::OneOf        => 'FilterOneOf',
			Filter::RangeFrom    => 'FilterRangeFrom',
			Filter::RangeTo      => 'FilterRangeTo',
			Filter::SearchString => 'FilterSearchString'
		);
		
		private static function createFilterParam($matchType, $paramName) {
			if (isset(self::$filterClassNames[$matchType])) {
				$className = '\\Framework\\' . self::$filterClassNames[$matchType];
				return new $className($paramName);
			}
			return null;
		}
		
		public static function create(array $filterParams, array $searchCols) {
			$searchString = null;
			$params = array();
			
			foreach ($filterParams as $paramName => $matchType) {
				if (isset($_GET[$paramName])) {
					if ($matchType == Filter::SearchString && count($searchCols) > 0) {
						$searchString = new FilterSearchString($paramName, $searchCols);
 					} elseif ($param = self::createFilterParam($matchType, $paramName)) {
						$params[$paramName] = $param;
					}
				}
			}
			
			if (count($params) > 0 || $searchString !== null) {
				return new Filter($params, $searchString);
			}

			return null;
		}
		
		public static function apply(Filter $filter = null, array $columnNames) {
			$queryCondition = '';
			$paramArray = array();
			$paramTypes = '';
			
			if ($filter) {
				foreach ($columnNames as $paramName => $columnName) {
					if ($filterArg = $filter->__get($paramName)) {
						$queryCondition .= $filterArg->sqlCondition($columnName);
						$paramTypes .= $filterArg->sqlParamTypes();
						$paramArray = array_merge($paramArray, $filterArg->sqlParameters());
					}
				}
				
				$searchString = $filter->__searchString;
				
				if ($searchString !== null) {
					$queryCondition .= $searchString->sqlCondition($columnNames);
					$paramTypes .= $searchString->sqlParamTypes();
					$paramArray = array_merge($paramArray, $searchString->sqlParameters());
				}
			}
			
			return array($queryCondition, $paramTypes, $paramArray);
			
		}

		public function __get($name) {
			$name = ucfirst($name);

			if (isset($this->__params[$name])) {
				return $this->__params[$name];
			} else {
				return NULL;
			}
		}

		private function __construct(array $params, FilterSearchString $searchString = null) {
			$this->__params = $params;
			$this->__searchString = $searchString;
		}

		private $__params;
		private $__searchString;
	}
}

?>
