<?php

namespace Repositories {

	use Framework\Repository;
	use Framework\PageInfo;
	use Framework\SortInfo;
	use Framework\Filter;

	use Models\Alert;
	use Models\Wgs84Coordinates;

	class AlertRepository extends Repository {
		private static $sortColumnNames = array(
			'DateTime' => 'DateTime',
			'Name' => 'Name',
			'IventoryName' => 'InventoryName',
			'Severity' => 'FIELD(Severity, \'info\', \'warning\', \'critical\')',
			'Status' => 'FIELD(Status, \'unknown\', \'closed\', \'inactive\', \'active\')'
		);
		
		private function _composeQuery($columnList, $tableList, $condition = 'TRUE') {
			$query = 'SELECT '
				. $columnList
				. ' FROM '
				. $tableList;
				
			if ($this->restrictedUser && $this->userId !== null) {
				$query .= ' JOIN User2Network ON User2Network.userId = ' 
					. $this->userId 
					. ' AND User2Network.networkId = InventoryNetworkId';
			}

			$query .= ' WHERE ' . $condition;
			
			return $query;
		}

		private static $queryColumns = 'Id, DefinitionId, InventoryId, Name, DateTime, Severity, Status, Cleared, Message, Value, SecondaryValue, InventoryName, InventoryNetworkId, InventoryNetworkName, InventoryIPAddress, InventoryWarrantyStatus, InventoryCoordinatesLongitude, InventoryCoordinatesLatitude, TodoId, TodoUserId, TodoUserFullName, TodoStatus';
		private static $queryTables = 'ApiAlertsView';

		private function baseQuery() {
			return  $this->_composeQuery(self::$queryColumns, self::$queryTables);
		}
		
		private function countQuery() {
			 return $this->_composeQuery('COUNT(*)', self::$queryTables);
		}

		public function getById($itemId) {
			$query = $this->baseQuery() . ' AND Id = ?';

			$statement = $this->dataBase->prepare($query);
			$statement->bind_param('i', $itemId);
			$statement->execute();

			$alerts = $this->getItemsFromStatement($statement);

			$statement->close();

			if (count($alerts) != 1) {
				return null;
			}

			return $alerts[0];
		}
		
		private static $filterColumnNames = array(
			'DateFrom' => 'DATE(DateTime)',
			'DateTo'   => 'DATE(DateTime)',
			'NameContains' => 'Name',
			'MessageContains' => 'Message',
			'InventoryNameContains' => 'InventoryName',
			'InventoryIPAddressContains' => 'InventoryIPAddress',
			'InventoryNetworkNameContains' => 'InventoryNetworkName',
			'TodoUserFullNameContains' => 'TodoUserFullName',
			'InventoryNetworkId' => 'InventoryNetworkId',
			'Name' => 'Name',
			'Message' => 'Message',
			'InventoryName' => 'InventoryName',
			'InventoryIPAddress' => 'InventoryIPAddress',
			'InventoryNetworkName' => 'InventoryNetworkName',
			'TodoUserFullName' => 'TodoUserFullName',
			'Severity' => 'Severity',
			'SeverityOneOf' => 'Severity',
			'Status' => 'Status'
		);
			
		public function getList(Filter $filter = null, PageInfo $pageInfo = null, SortInfo $sortInfo = null) {
			$query = $this->baseQuery();
			
			list($queryCondition, $paramTypes, $paramArray) = Filter::apply($filter, self::$filterColumnNames);
			
			/* Hack: a special case for the Status=active filter */
			if (isset($_GET['Status']) && $_GET['Status'] === 'active') {
				$queryCondition .= ' AND ApiAlertsView.Id IN (SELECT alarm_id FROM ActiveAlarm)';
			}
			
			$query .= $queryCondition;
			

			if ($sortInfo) {
				$query .= $sortInfo->sqlCode(self::$sortColumnNames);
			}

			if ($pageInfo) {
				$query .= $pageInfo->sqlCode();

				$countQuery = $this->countQuery() . $queryCondition;
				$countStmt = $this->dataBase->prepare($countQuery);

				if (count($paramArray) > 0) {
					call_user_func_array(
						array($countStmt, 'bind_param'),
						array_merge(array($paramTypes), $paramArray)
					);
				}

				$countStmt->execute();
				$countStmt->bind_result($itemCount);
				$countStmt->fetch();
				$countStmt->close();

				$pageInfo->setTotalItems($itemCount);
			}
			
			$statement = $this->dataBase->prepare($query);
			
			

			if (count($paramArray) > 0) {
				call_user_func_array(array($statement, 'bind_param'), array_merge(array($paramTypes), $paramArray));
			}

			$statement->execute();

			$alerts = $this->getItemsFromStatement($statement);

			$statement->close();

			return $alerts;
		}
		
		private static $groupByColumns = array(
			'InventoryNetworkId' => 'Networks.id'
		);

		private static $groupByTables = array(
			'InventoryNetworkId' => 'Networks LEFT JOIN ApiAlertsView ON InventoryNetworkId = Networks.id'
		);

		private static $groupBySort = array(
			'InventoryNetworkId' => ' ORDER BY Networks.name'
		);
		
		public function getCount(Filter $filter = null, $groupBy = null) {
			
			if ($groupBy != null && !isset(self::$groupByColumns[$groupBy])) {
				return null;
			}
			
			if ($groupBy != null) {
				$groupByColumn = self::$groupByColumns[$groupBy];
			}
			
			$queryColumns = 'COUNT(ApiAlertsView.Id)';
			
			if ($groupBy != null) {
				$queryColumns .= ', ' . $groupByColumn;
			}
			
			$tables = self::$queryTables;

			if ($groupBy != null && isset(self::$groupByTables[$groupBy])) {
				$tables = self::$groupByTables[$groupBy];
			}

			$query = $this->_composeQuery($queryColumns, $tables);

			list($queryCondition, $paramTypes, $paramArray) = Filter::apply($filter, self::$filterColumnNames);
			
			/* Hack: a special case for the Status=active filter */
			if (isset($_GET['Status']) && $_GET['Status'] === 'active') {
				$queryCondition .= ' AND ApiAlertsView.Id IN (SELECT alarm_id FROM ActiveAlarm)';
			}
			
			$query .= $queryCondition;
			
			if ($groupBy != null) {
				$query .= ' GROUP BY ' . $groupByColumn;
			}

			if ($groupBy != null && isset(self::$groupBySort[$groupBy])) {
				$query .= self::$groupBySort[$groupBy];
			}
			
			$statement = $this->dataBase->prepare($query);

			if (count($paramArray) > 0) {
				call_user_func_array(array($statement, 'bind_param'), array_merge(array($paramTypes), $paramArray));
			}

			$statement->execute();
			
			if ($groupBy == null) {
				$statement->bind_result($alertCount);
				$statement->fetch();
				$statement->close();
				
				$count = new \Models\AlertCount;
				$count->AlertCount = $alertCount;
				
				return $count;
			}
			
			$statement->bind_result($alertCount, $groupByResult);
			
			$results = array();
			
			while ($statement->fetch()) {
				$count = new \Models\AlertCount;
				
				$count->AlertCount = $alertCount;
				$count->$groupBy = is_null($groupByResult) ? null : (string) $groupByResult;
				
				$results[] = $count;
			}
			
			$statement->close();

			return $results;
		}

		protected function getItemsFromStatement(\mysqli_stmt $statement) {
			$statement->bind_result(
				$id,
				$definitionId,
				$inventoryId,
				$name,
				$dateTime,
				$severity,
				$status,
				$cleared,
				$message,
				$value,
				$secondaryValue,
				$inventoryName,
				$inventoryNetworkId,
				$inventoryNetworkName,
				$inventoryIPAddress,
				$inventoryWarrantyStatus,
				$inventoryLongitude,
				$inventoryLatitude,
				$todoId,
				$todoUserId,
				$todoUserFullName,
				$todoStatus
			);

			$alerts = array();

			while ($statement->fetch()) {
				$alert = new Alert;

				$alert->Id = (string) $id;
				$alert->DefinitionId = (string) $definitionId;
				$alert->InventoryId = (string) $inventoryId;
				$alert->Name = $name;
				$alert->DateTime = $dateTime;
				$alert->Severity = $severity;
				$alert->Status = $status;
				$alert->Cleared = ($cleared != 0);
				$alert->Message = $message;
				$alert->Value = (double) $value;
				$alert->SecondaryValue = (double) $secondaryValue;
				$alert->InventoryName = $inventoryName;
				$alert->InventoryNetworkId = (string) $inventoryNetworkId;
				$alert->InventoryNetworkName = $inventoryNetworkName;
				$alert->InventoryIPAddress = $inventoryIPAddress;
				$alert->InventoryWarrantyStatus = $inventoryWarrantyStatus;
				$alert->TodoId = $todoId;
				$alert->TodoUserId = $todoUserId;
				$alert->TodoUserFullName = $todoUserFullName;
				$alert->TodoStatus = $todoStatus;
				
				$alert->InventoryCoordinates = new Wgs84Coordinates;
				$alert->InventoryCoordinates->Longitude = $inventoryLongitude;
				$alert->InventoryCoordinates->Latitude = $inventoryLatitude;

				$alerts[] = $alert;
			}

			return $alerts;
		}
	}
}

?>
