<?php
namespace Repositories;
	use Framework\Repository;
	use Framework\PageInfo;
	use Framework\SortInfo;
	use Framework\Filter;
	
	use Models\InventoryCompact;
	use Models\Inventory;
	use Models\Wgs84Coordinates;
	
	class InventoryRepository extends Repository {
		private static $sortColumnNames = array(
			'Name' => 'Hosts.name',
			'IPAddress' => 'INET_NTOA(Hosts.ipaddr)',
			'NetworkName' => 'Networks.name',
			'DeviceType' => 'DeviceTypes.name',
			'IsRegistered' => 'Hosts.registered',
			'IsOffline' => 'Hosts.isDown'
		);
	
		private static $itemColumns = 'Hosts.id, Hosts.name, INET_NTOA(Hosts.ipaddr), Hosts.networkId, Hosts.deviceTypeId, Hosts.registered, Hosts.isDown, Networks.name, DeviceTypes.name, Hosts.hbInterval, Hosts.mac, Hosts.serialno, Hosts.location, Hosts.firmware, Hosts.lastContact, Hosts.alarmProfileId, Hosts.statsProfileId, IFNULL(IF(Hosts.latitude_local = 0 AND Hosts.longitude_local = 0, NULL, Hosts.latitude_local), Hosts.latitude), IFNULL(IF(Hosts.latitude_local = 0 AND Hosts.longitude_local = 0, NULL, Hosts.longitude_local), Hosts.longitude), AlarmProfiles.name, StatsProfiles.name, IFNULL(Hosts.height_agl_local, Hosts.height_agl)';
		private static $itemTables = 'Hosts LEFT JOIN Networks ON Hosts.networkId = Networks.id LEFT JOIN DeviceTypes ON Hosts.deviceTypeId = DeviceTypes.id LEFT JOIN AlarmProfiles ON Hosts.alarmProfileId = AlarmProfiles.id LEFT JOIN StatsProfiles ON Hosts.statsProfileId = StatsProfiles.id';

		public function getById($itemId) {
			$itemQuery = $this->composeQuery(self::$itemColumns, self::$itemTables, 'Hosts.id = ?');
			
			$statement = $this->dataBase->prepare($itemQuery);
			$statement->bind_param('i', $itemId);
			$statement->execute();
			
			$statement->bind_result(
				$deviceId,
				$deviceName,
				$deviceIPAddress,
				$deviceNetworkId,
				$deviceTypeId,
				$deviceIsRegistered,
				$deviceIsOffline,
				$deviceNetworkName,
				$deviceTypeName,
				$deviceHeartbeatInterval,
				$deviceMacAddress,
				$deviceSerialNumber,
				$deviceLocation,
				$deviceFirmware,
				$deviceLastHeartbeat,
				$deviceAlertProfileId,
				$deviceStatsProfileId,
				$deviceLatitude,
				$deviceLongitude,
				$deviceAlertProfileName,
				$deviceStatsProfileName,
				$deviceHeight
			);
			
			$found = $statement->fetch();
			
			$statement->close();
			
			if (!$found) {
				return null;
			}
			
			$device = new Inventory;
			
			$device->Id = (string) $deviceId;
			$device->Name = $deviceName;
			$device->IPAddress = $deviceIPAddress;
			$device->NetworkId = is_null($deviceNetworkId) ? null : (string) $deviceNetworkId;
			$device->DeviceTypeId = is_null($deviceTypeId) ? null : (string) $deviceTypeId;
			$device->IsRegistered = ($deviceIsRegistered != 0);
			$device->IsOffline = ($deviceIsOffline != 0);
			$device->NetworkName = $deviceNetworkName;
			$device->DeviceType = $deviceTypeName;
			$device->HeartbeatInterval = $deviceHeartbeatInterval;
			$device->MacAddress = $deviceMacAddress;
			$device->SerialNumber = $deviceSerialNumber;
			$device->Location = $deviceLocation;
			$device->Firmware = $deviceFirmware;
			$device->LastHeartbeat = $deviceLastHeartbeat;
			$device->AlertProfileId = is_null($deviceAlertProfileId) ? null : (string) $deviceAlertProfileId;
			$device->StatsProfileId = is_null($deviceStatsProfileId) ? null : (string) $deviceStatsProfileId;
			
			$coordinates = new Wgs84Coordinates;
			
			$coordinates->Longitude = is_null($deviceLongitude) ? null : (float) $deviceLongitude;
			$coordinates->Latitude = is_null($deviceLatitude) ? null : (float) $deviceLatitude;
			
			$device->Coordinates = $coordinates;
			
			$device->AlertProfileName = $deviceAlertProfileName;
			$device->StatsProfileName = $deviceStatsProfileName;
			
			$device->Height = is_null($deviceHeight) ? null : (float) $deviceHeight;
			
			return $device;
		}
		
		private static $filterColumnNames = array(
			'NameContains' => 'Hosts.name',
			'IPAddressContains' => 'INET_NTOA(Hosts.ipaddr)',
			'NetworkNameContains' => 'Networks.name',
			'DeviceTypeOneOf' => 'DeviceTypes.name',
			'NetworkId' => 'Networks.id',
			'Name' => 'Hosts.name',
			'IPAddress' => 'INET_NTOA(Hosts.ipaddr)',
			'NetworkName' => 'Networks.name',
			'DeviceType' => 'DeviceTypes.name',
			'Location' => 'Hosts.location',
			'MacAddress' => 'Hosts.mac',
			'Firmware' => 'Hosts.firmware',
			'IsRegistered' => 'Hosts.registered',
			'IsOffline' => 'Hosts.isDown',
			'SerialNumber' => 'Hosts.serialno'
		);
		
		private static $groupByColumns = array(
			'NetworkId' => 'Networks.id'
		);
		
		private static $groupByOrder = array(
			'NetworkId' => 'Networks.name'
		);

		public function getCount(Filter $filter = null, $groupBy = null) {
			$groupByColumn = null;
			
			if ($groupBy != null && isset(self::$groupByColumns[$groupBy])) {
				$groupByColumn = self::$groupByColumns[$groupBy];
			}
			
			if ($groupBy != null && $groupByColumn == null) {
				return null;
			}
			
			$queryColumns = 'COUNT(Hosts.id)';
			
			if ($groupByColumn != null) {
				$queryColumns .= ', ' . $groupByColumn;
			}
			
			$queryTables = self::$listTables;
            
			if ($groupBy != null && isset(self::$groupByTables[$groupBy])) {
				$queryTables = self::$groupByTables[$groupBy];
			}
			
			$query = $this->composeQuery($queryColumns, $queryTables);
			
			list($queryCondition, $paramTypes, $paramArray) = Filter::apply($filter, self::$filterColumnNames);
			
			$query .= $queryCondition;
			
			if ($groupByColumn != null) {
				$query .= ' GROUP BY ' . $groupByColumn;
			}
			
			if ($groupBy != null && isset(self::$groupByOrder[$groupBy])) {
				$query .= ' ORDER BY ' . self::$groupByOrder[$groupBy];
			}

			$statement = $this->dataBase->prepare($query);

			if (count($paramArray) > 0) {
				call_user_func_array(array($statement, 'bind_param'), array_merge(array($paramTypes), $paramArray));
			}

			$statement->execute();
			
			if ($groupByColumn == null) {
				$statement->bind_result($inventoryCount);
				$statement->fetch();
				$statement->close();
				
				$count = new \Models\InventoryCount;
				$count->InventoryCount = $inventoryCount;
				
				return $count;
			}
			
			$statement->bind_result($inventoryCount, $groupByResult);
			
			$results = array();
			
			while ($statement->fetch()) {
				$count = new \Models\InventoryCount;
				
				$count->InventoryCount = $inventoryCount;
				$count->$groupBy = is_null($groupByResult) ? null : (string) $groupByResult;
				
				$results[] = $count;
			}
			
			$statement->close();


			if ($groupByColumn != null) {
				$query = $this->composeQuery($queryColumns, self::$listTables);
				$query .= $queryCondition . ' AND ' . $groupByColumn . ' IS NULL';


				$statement = $this->dataBase->prepare($query);

				if (count($paramArray) > 0) {
					call_user_func_array(array($statement, 'bind_param'), array_merge(array($paramTypes), $paramArray));
				}
				
				$statement->execute();

				$statement->bind_result($inventoryCount, $groupByResult);
				$statement->fetch();

				$count = new \Models\InventoryCount;
				
				$count->InventoryCount = $inventoryCount;
				$count->$groupBy = null;

				$results[] = $count;

				$statement->close();

			}

			return $results;
			
		
		}

		public function getList(Filter $filter = null, PageInfo $pageInfo = null, SortInfo $sortInfo = null) {
			$query = $this->listQuery();
			
			list ($queryCondition, $paramTypes, $paramArray) = Filter::apply($filter, self::$filterColumnNames);

			$query .= $queryCondition;
			
			if ($sortInfo) {
				$query .= $sortInfo->sqlCode(self::$sortColumnNames);
			}
			
			if ($pageInfo) {
				$query .= $pageInfo->sqlCode();
				
				$countQuery = $this->countQuery() . $queryCondition;
				$countStmt = $this->dataBase->prepare($countQuery);

				if (count($paramArray) > 0) {
					call_user_func_array(
						array($countStmt, 'bind_param'),
						array_merge(array($paramTypes), $paramArray)
					);
				}

				$countStmt->execute();
				$countStmt->bind_result($itemCount);
				$countStmt->fetch();
				$countStmt->close();

				$pageInfo->setTotalItems($itemCount);
			}
			
			$statement = $this->dataBase->prepare($query);
			
			if (count($paramArray) > 0) {
				call_user_func_array(array($statement, 'bind_param'), array_merge(array($paramTypes), $paramArray));
			}
			
			$statement->execute();
			
			$inventory = $this->getItemsFromStatement($statement);
			
			$statement->close();
			
			return $inventory;
		}
		
		private static $listColumns = 'Hosts.id, Hosts.name, INET_NTOA(Hosts.ipaddr), Hosts.networkId, Hosts.deviceTypeId, Hosts.registered, Hosts.isDown, Networks.name, DeviceTypes.name, IFNULL(IF(Hosts.latitude_local = 0 AND Hosts.longitude_local = 0, NULL, Hosts.latitude_local), Hosts.latitude), IFNULL(IF(Hosts.latitude_local = 0 AND Hosts.longitude_local = 0, NULL, Hosts.longitude_local), Hosts.latitude), IFNULL(Hosts.height_agl_local, Hosts.height_agl)';
		private static $listTables = 'Hosts LEFT JOIN Networks ON Hosts.networkId = Networks.id LEFT JOIN DeviceTypes ON Hosts.deviceTypeId = DeviceTypes.id';
		
		private static $groupByTables = array(
			'NetworkId' => 'Networks LEFT JOIN Hosts ON Hosts.networkId = Networks.id LEFT JOIN DeviceTypes ON Hosts.deviceTypeId = DeviceTypes.id'
		);
		
		private function listQuery() {
			return $this->composeQuery(self::$listColumns, self::$listTables);
		}
		
		private function countQuery() {
			return $this->composeQuery('COUNT(*)', self::$listTables);
		}
		
		protected function getItemsFromStatement($statement) {
			$inventory = array();
			
			$statement->bind_result(
				$deviceId,
				$deviceName,
				$deviceIPAddress,
				$deviceNetworkId,
				$deviceTypeId,
				$deviceIsRegistered,
				$deviceIsOffline,
				$deviceNetworkName,
				$deviceTypeName,
				$deviceLatitude,
				$deviceLongitude,
				$deviceHeight
			);
			
			while ($statement->fetch()) {
				$device = new InventoryCompact;
				
				$device->Id = (string) $deviceId;
				$device->Name = $deviceName;
				$device->IPAddress = $deviceIPAddress;
				$device->NetworkId = is_null($deviceNetworkId) ? null : (string) $deviceNetworkId;
				$device->DeviceTypeId = is_null($deviceTypeId) ? null : (string) $deviceTypeId;
				$device->IsRegistered = ($deviceIsRegistered != 0);
				$device->IsOffline = ($deviceIsOffline != 0);
				$device->NetworkName = $deviceNetworkName;
				$device->DeviceType = $deviceTypeName;
			
				$coordinates = new Wgs84Coordinates;
				
				$coordinates->Longitude = is_null($deviceLongitude) ? null : (float) $deviceLongitude;
				$coordinates->Latitude = is_null($deviceLatitude) ? null : (float) $deviceLatitude;
				
				$device->Coordinates = $coordinates;
				
				$device->Height = is_null($deviceHeight) ? null : (float) $deviceHeight;

				$inventory[] = $device;
			}
			
			return $inventory;
		}
		
		public function delete($inventoryId) {
			$statement = $this->dataBase->prepare('DELETE FROM Hosts WHERE id = ?');
			$statement->bind_param('i', $inventoryId);
			$statement->execute();
			$statement->close();
		}
		
		public function update(Inventory $inventory) {
			$database = $this->dataBase;

			$database->autocommit(false);
	    
			$database->query('DELETE FROM Host2Stats WHERE hostId=' . $inventory->Id);
			
			if (!is_null($inventory->StatsProfileId)) {
				$database->query('INSERT INTO Host2Stats (hostId, statsId) SELECT '
					. $inventory->Id
					. ', statsDefId FROM StatsProfile2StatsDef WHERE statsProfileId='
					. $inventory->StatsProfileId
				);
			}
			
			$database->query('DELETE FROM Host2Alarm WHERE hostId=' . $inventory->Id);

			if (!is_null($inventory->AlertProfileId)) {
				$database->query('INSERT INTO Host2Alarm (hostId, alarmId) SELECT '
					. $inventory->Id 
					. ', alarmId FROM AlarmProfile2AlarmDef WHERE alarmProfileId='
					. $inventory->AlertProfileId
				);
			}
			
			$update = 'UPDATE Hosts SET networkId=?, alarmProfileId=?, statsProfileId=?, deviceTypeId=?, effectiveHBInterval = IFNULL(effectiveHbInterval, NULLIF(hbInterval, ?)), hbInterval=?, registered=?, longitude_local=?, latitude_local=? WHERE id = ?';
			
			$statement = $database->prepare($update);
			$statement->bind_param('iiiiiiiddi',
				$inventory->NetworkId,
				$inventory->AlertProfileId,
				$inventory->StatsProfileId,
				$inventory->DeviceTypeId,
				$inventory->HeartbeatInterval,
				$inventory->HeartbeatInterval,
				$inventory->IsRegistered,
				$inventory->Coordinates->Longitude,
				$inventory->Coordinates->Latitude,
				$inventory->Id
			);

			$statement->execute();
			
			
			// If they have supplied the height property, do update it
			 
			if ($inventory->Height !== FALSE) {
				$heightUpdate = 'UPDATE Hosts SET height_agl_local=? WHERE id=?';
				$heightStmt = $database->prepare($heightUpdate);
				$heightStmt->bind_param('di', 
					$inventory->Height, 
					$inventory->Id
				);
				$heightStmt->execute();
			}
			
			$database->commit();

			$statement->close();
			
			if (isset($heightStmt)) {
				$heightStmt->close();
			}
		}
	}

	
