<?php

namespace Repositories {
	
	use \Framework\Repository;
	use \Framework\Filter;
	use \Framework\PageInfo;
	use \Framework\SortInfo;
	
	use \Models\Todo;
	use \Models\TodoNote;
	
	class TodoRepository extends Repository {
		private static $sortColumnNames = array(
			'DateCreated' => 'Todos.dateCreated',
			'Priority' => 'FIELD(Todos.priority, \'low\', \'medium\', \'high\')',
			'Status' => 'FIELD(Todos.status, \'unassigned\', \'assigned\', \'cannotComplete\', \'completed\')'
		);
		
		private static $filterColumnNames = array(
			'InventoryNetworkId' => 'Networks.id',
			'Name' => 'Todos.name',
			'InventoryIPAddress' => 'INET_NTOA(Hosts.ipaddr)',
			'InventoryName' => 'Hosts.name',
			'InventoryNetworkName' => 'Networks.name',
			'UserFullName' => 'CONCAT(Users.firstname, \' \', Users.lastname)',
			'AlertName' => 'AlarmDefinition.name',
			'InventoryLocation' => 'Hosts.location',
			'Status' => 'Todos.status',
			'Priority' => 'Todos.priority',
			'DateCreatedFrom' => 'DATE(Todos.dateCreated)',
			'DateCreatedTo' => 'DATE(Todos.dateCreated)',
			'StatusOneOf' => 'Todos.status',
			'PriorityOneOf' => 'Todos.priority',
			'UserId' => 'Todos.userId'
			
		);
		
		private static $groupByColumns = array(
			'InventoryNetworkId' => 'Networks.id'
		);

		private static $groupByTables = array(
			'InventoryNetworkId' => 'Networks LEFT JOIN Hosts ON Hosts.networkId = Networks.id LEFT JOIN Alarms ON Alarms.hostId = Hosts.id LEFT JOIN Todos ON Todos.alarmId = Alarms.id LEFT JOIN AlarmDefinition ON Alarms.alarmId = AlarmDefinition.id LEFT JOIN Users ON Users.id = Todos.userId LEFT JOIN HostCustomerServicesStatusView ON Hosts.id = HostCustomerServicesStatusView.hostId'
		);

		private static $groupByOrder = array(
			'InventoryNetworkId' => 'Networks.name'
		);
	
		private static $queryColumns = 'Todos.id, Todos.name, Todos.createdBy, Todos.dateCreated, Todos.priority, Todos.status, Todos.alarmId, Todos.userId, Hosts.id, Hosts.name, IF(HostCustomerServicesStatusView.hostId IS NULL, NULL, HostCustomerServicesStatusView.status), INET_NTOA(Hosts.ipaddr), Networks.id, Networks.name, CONCAT(Users.firstname, \' \', Users.lastname), AlarmDefinition.name, IF(Hosts.latitude_local = 0 AND Hosts.longitude_local = 0, NULL, Hosts.latitude_local), IF(Hosts.latitude_local = 0 AND Hosts.longitude_local = 0, NULL, Hosts.longitude_local)';

		private static $queryTables = <<<EOQ
			Todos
			INNER JOIN Alarms ON Todos.alarmId = Alarms.id
			INNER JOIN Hosts ON Alarms.hostId = Hosts.id
			INNER JOIN Networks ON Hosts.networkId = Networks.id
			INNER JOIN AlarmDefinition ON Alarms.alarmId = AlarmDefinition.id
			LEFT JOIN Users ON Users.id = Todos.userId
			LEFT JOIN HostCustomerServicesStatusView ON Hosts.id = HostCustomerServicesStatusView.hostId
EOQ;

		public function getById($itemId) {
			$query = $this->composeQuery(self::$queryColumns, self::$queryTables, 'Todos.id = ?');
			
			$statement = $this->dataBase->prepare($query);
			
			$statement->bind_param('i', $itemId);
			$statement->execute();
			
			$todos = $this->getItemsFromStatement($statement);
			
			$statement->close();
			
			if (count($todos) != 1) {
				return null;
			}
			
			$todo = $todos[0];
			
			$todo->Notes = $this->getNotesList($todo->Id);
			
			return $todo;
		}
		
		public function getList(Filter $filter = null, PageInfo $pageInfo = null, SortInfo $sortInfo = null) {
			$query = $this->composeQuery(self::$queryColumns, self::$queryTables);
			
			list ($queryCondition, $paramTypes, $paramArray) = Filter::apply($filter, self::$filterColumnNames);

			$query .= $queryCondition;
			
			if ($sortInfo) {
				$query .= $sortInfo->sqlCode(self::$sortColumnNames);
			}

			if ($pageInfo) {
				$query .= $pageInfo->sqlCode();

				$countQuery = self::countQuery() . $queryCondition;
				$countStmt = $this->dataBase->prepare($countQuery);

				if (count($paramArray) > 0) {
					call_user_func_array(
						array($countStmt, 'bind_param'),
						array_merge(array($paramTypes), $paramArray)
					);
				}

				$countStmt->execute();
				$countStmt->bind_result($itemCount);
				$countStmt->fetch();
				$countStmt->close();

				$pageInfo->setTotalItems($itemCount);
			}
			
			$statement = $this->dataBase->prepare($query);
			
			if (count($paramArray) > 0) {
				call_user_func_array(array($statement, 'bind_param'), array_merge(array($paramTypes), $paramArray));
			}

			$statement->execute();
			
			$todos = $this->getItemsFromStatement($statement);
			
			$statement->close();
			
			foreach ($todos as $todo) {
				$todo->Notes = $this->getNotesList($todo->Id);
			}
			
			return $todos;
		}
		
		public function getCount(Filter $filter = null, $groupBy = null) {
			$groupByColumn = null;
			
			if ($groupBy != null && isset(self::$groupByColumns[$groupBy])) {
				$groupByColumn = self::$groupByColumns[$groupBy];
			}
			
			if ($groupBy != null && $groupByColumn == null) {
				return null;
			}
			
			$queryColumns = 'COUNT(Todos.Id)';
			
			if ($groupByColumn != null) {
				$queryColumns .= ', ' . $groupByColumn;
			}

			$queryTables = self::$queryTables;

			if ($groupBy != null && isset(self::$groupByTables[$groupBy])) {
				$queryTables = self::$groupByTables[$groupBy];
			}
			
			$query = 'SELECT ' . $queryColumns . ' FROM ' . $queryTables . ' WHERE TRUE';
			
			list($queryCondition, $paramTypes, $paramArray) = Filter::apply($filter, self::$filterColumnNames);
			
			$query .= $queryCondition;
			
			if ($groupByColumn != null) {
				$query .= ' GROUP BY ' . $groupByColumn;
			}

			if ($groupBy != null && isset(self::$groupByOrder[$groupBy])) {
				$query .= ' ORDER BY ' . self::$groupByOrder[$groupBy];
			}

			$statement = $this->dataBase->prepare($query);

			if (count($paramArray) > 0) {
				call_user_func_array(array($statement, 'bind_param'), array_merge(array($paramTypes), $paramArray));
			}

			$statement->execute();
			
			if ($groupByColumn == null) {
				$statement->bind_result($inventoryCount);
				$statement->fetch();
				$statement->close();
				
				$count = new \Models\TodoCount;
				$count->TodoCount = $inventoryCount;
				
				return $count;
			}
			
			$statement->bind_result($inventoryCount, $groupByResult);
			
			$results = array();
			
			while ($statement->fetch()) {
				$count = new \Models\TodoCount;
				
				$count->TodoCount = $inventoryCount;
				$count->$groupBy = is_null($groupByResult) ? null : (string) $groupByResult;
				
				$results[] = $count;
			}
			
			$statement->close();

			return $results;
		}
		
		public function create(Todo $item) {
			$insert = 'INSERT INTO Todos (createdBy, dateCreated, name, priority, status, alarmId, userId) VALUES (?, UTC_TIMESTAMP(), ?, ?, ?, ?, ?)';
			
			$statement = $this->dataBase->prepare($insert);
			
			$statement->bind_param('isssii',
				$item->CreatedBy,
				$item->Name,
				$item->Priority,
				$item->Status,
				$item->AlertId,
				$item->UserId
			);
			
			$statement->execute();
			
			$item->Id = $statement->insert_id;
			
			$statement->close();
			
			foreach ($item->Notes as $note) {
				$note->UserId = $item->UserId;
				$note->TodoId = $item->Id;
				$this->createNote($note);
			}
		}
		
		public function update(Todo $item) {
			$update = 'UPDATE Todos SET updatedBy = ?, name = ?, priority = ?, status = ?, userId = ? WHERE id = ?';
			
			$statement = $this->dataBase->prepare($update);
			$statement->bind_param('isssii', $this->userId, $item->Name, $item->Priority, $item->Status, $item->userId, $item->Id);
			$statement->execute();
			$statement->close();
		}
		
		public function delete($todoId) {
			$delete = 'DELETE FROM Todos WHERE id = ?';
			
			$statement = $this->dataBase->prepare($delete);
			$statement->bind_param('i', $todoId);
			$statement->execute();
			$statement->close();
		}
		
		public function getNotesList($todoId) {
			$query = 'SELECT TodoNotes.id, todoId, userId, dateCreated, dateEdited, note, CONCAT(firstname, \' \', lastname) FROM TodoNotes, Users WHERE userId = Users.Id AND todoId = ? ORDER BY dateCreated';
			
			$statement = $this->dataBase->prepare($query);
			$statement->bind_param('i', $todoId);
			$statement->execute();
			
			$notes = $this->getNotesFromStatement($statement);
			
			$statement->close();
			
			return $notes;
		}
		
		public function createNote(TodoNote $note) {
			$insert = 'INSERT INTO TodoNotes (todoId, userId, note) VALUES (?, ? ,?)';
			
			$statement = $this->dataBase->prepare($insert);
			$statement->bind_param('iis', $note->TodoId, $note->UserId, $note->Note);
			$statement->execute();
			
			$note->Id = $statement->insert_id;
			
			$statement->close();
		}
		
		protected function getItemsFromStatement(\mysqli_stmt $statement) {
			$statement->bind_result(
				$todoId,
				$todoName,
				$todoCreatedBy,
				$todoDateCreated,
				$todoPriority,
				$todoStatus,
				$todoAlertId,
				$todoUserId,
				$todoInventoryId,
				$todoInventoryName,
				$todoInventoryWarrantyStatus,
				$todoInventoryIPAddress,
				$todoInventoryNetworkId,
				$todoInventoryNetworkName,
				$todoUserFullName,
				$todoAlertName,
				$todoInventoryCoordsLatitude,
				$todoInventoryCoordsLongitude
			);
			
			$items = array();
			
			while ($statement->fetch()) {
				$todo = new Todo;
				
				$todo->Id = $todoId;
				$todo->Name = $todoName;
				$todo->CreatedBy = (string) $todoCreatedBy;
				$todo->DateCreated = $todoDateCreated;
				$todo->Priority = $todoPriority;
				$todo->Status = $todoStatus;
				$todo->AlertId = (string) $todoAlertId;
				$todo->UserId = is_null($todoUserId) ? null : (string) $todoUserId;
				$todo->InventoryId = (string) $todoInventoryId;
				$todo->InventoryName = $todoInventoryName;
				$todo->InventoryWarrantyStatus = $todoInventoryWarrantyStatus;
				$todo->InventoryIPAddress = $todoInventoryIPAddress;
				$todo->InventoryNetworkId = (string) $todoInventoryNetworkId;
				$todo->InventoryNetworkName = $todoInventoryNetworkName;
				$todo->AlertName = $todoAlertName;
				$todo->UserFullName = $todoUserFullName;
				
				$coordinates = new \Models\Wgs84Coordinates;
				
				$coordinates->Longitude = (float) $todoInventoryCoordsLongitude;
				$coordinates->Latitude = (float) $todoInventoryCoordsLatitude;
				
				$todo->InventoryCoordinates = $coordinates;
				
				
				$items[] = $todo;
			}
			
			return $items;
		}
		
		private function getNotesFromStatement(\mysqli_stmt $statement) {
			$notes = array();
			
			$statement->bind_result(
				$noteId,
				$todoId,
				$userId,
				$dateCreated,
				$dateEdited,
				$noteText,
				$userFullName
			);
			
			while ($statement->fetch()) {
				$note = new TodoNote;
				
				$note->Id = $noteId;
				$note->TodoId = $todoId;
				$note->UserId = $userId;
				$note->DateCreated = $dateCreated;
				$note->DateEdited = $dateEdited;
				$note->Note = $noteText;
				$note->UserFullName = $userFullName;
				
				$notes[] = $note;
			}
			
			return $notes;
		}
		
		public function updateStatus($todoId, $status) {
			$query = 'UPDATE Todos SET updatedBy = ?, status = ? WHERE id = ?';
			
			$statement = $this->dataBase->prepare($query);
			$statement->bind_param('isi', $this->userId, $status, $todoId);
			$statement->execute();
			$statement->close();
		}
		
		public function updateUserId($todoId, $userId) {
			$query = 'UPDATE Todos SET updatedBy = ?, userId = ? WHERE id = ?';
			
			$statement = $this->dataBase->prepare($query);
			$statement->bind_param('iii', $this->userId, $userId, $todoId);
			$statement->execute();
			$statement->close();
		}
		
	}
}

?>
