<?php

use Framework\Filter;

class FilteredTodoResource extends Framework\Resource {
	protected function filterArgs() {
		return array(
			'InventoryNetworkId'	=> Filter::Equals,
			'DateCreatedFrom'	=> Filter::RangeFrom,
			'DateCreatedTo'		=> Filter::RangeTo,
			'PriorityOneOf'		=> Filter::OneOf,
			'StatusOneOf'		=> Filter::OneOf,
			'Priority'		=> Filter::Equals,
			'Status'		=> Filter::Equals,
			'UserId'		=> Filter::Equals,
			'SearchString'		=> Filter::SearchString
		);
	}
	
	protected function searchCols() {
		return array(
			'Name',
			'InventoryIPAddress',
			'InventoryName',
			'InventoryNetworkName',
			'UserFullName',
			'AlertName',
			'InventoryLocation',
			'Status',
			'Priority'
		);
	}
}

/**
 * @uri /todos
 */
class TodosCollection extends FilteredTodoResource {

	protected function sortCols() {
		return array(
			'DateCreated',
			'Priority',
			'Status'
		);
	}
	
	protected function defaultSortBy() {
		return 'DateCreated,Priority,Status';
	}
	
	protected function defaultSortOrder() {
		return 'D,D,D';
	}
	
	/**
	 * @method GET
	 * @provides application/json
	 */
	 public function getList() {
		$repository = new Repositories\TodoRepository($this->dataBase,
			$this->restrictedUser,
			$this->userId
			);
		$todoList = $repository->getList($this->filter, $this->pageInfo, $this->sortInfo);
		$response = new Tonic\Response(200, $todoList);
		return $response;
	 }
	 
	 /**
	  * @method POST
	  * @accepts application/json
	  */
	  public function create() {
		$data = $this->request->data;

		$todo = new Models\Todo;
		
		$todo->CreatedBy = $this->userId;
		$todo->Name = $data->Name;
		$todo->Priority = $data->Priority;
		$todo->Status = $data->Status;
		$todo->AlertId = $data->AlertId;
		$todo->UserId = $data->UserId;
		$todo->Notes = array();
		
		if (isset($data->Notes) && is_array($data->Notes)) {
			foreach($data->Notes as $reqNote) {
				$note = new Models\TodoNote;
				$note->UserId = $this->userId;
				$note->Note = $reqNote->Note;
				$todo->Notes[] = $note;
			}
		}

		$repository = new Repositories\TodoRepository($this->dataBase,
			$this->restrictedUser,
			$this->userId);
		
		$repository->create($todo);
		
		$response = new Tonic\Response(201);
		$response->Location = $this->baseUri . '/todos/' . $todo->Id;
		
		return $response;
		
	  }
}

/**
 * @uri /todos/count
 */
class TodoCountResource extends FilteredTodoResource {
	/**
	 * @method GET
	 * @provides application/json
	 */
	public function getCount() {
		$repository = new Repositories\TodoRepository($this->dataBase,
			$this->restrictedUser,
			$this->userId
			);
		$groupBy = isset($_GET['GroupBy']) ? $_GET['GroupBy'] : null;
		
		$count = $repository->getCount($this->filter, $groupBy);
		
		return $count == null
			? new Tonic\Response(400)
			: new Tonic\Response(200, $count);
	}
}

/**
 * @uri /todos/{todoId}
 */
class TodoResource extends Framework\Resource {
	/**
	 * @method GET
	 * @provides application/json
	 */
	public function get($todoId) {
		$repository = new Repositories\TodoRepository($this->dataBase,
			$this->restrictedUser,
			$this->userId
			);
		$todo = $repository->getById($todoId);
		if ($todo == null) {
			throw new Tonic\NotFoundException;
		}
		return new Tonic\Response(200, $todo);
	}
	
	/**
	 * @method POST
	 * @accepts application/json
	 */
	public function update($todoId) {
		$data = $this->request->data;
		
		$todo = new Models\Todo;
		
		$todo->Id = $todoId;
		$todo->Name = $data->Name;
		$todo->Priority = $data->Priority;
		$todo->Status = $data->Status;
		$todo->UserId = $data->UserId;
		
		$repository = new Repositories\TodoRepository($this->dataBase);
		$repository->update($todo);
		
		return new Tonic\Response(200);
	}
	
	/**
	 * @method DELETE
	 */
	public function delete($todoId) {
		$repository = new Repositories\TodoRepository($this->dataBase);
		$repository->delete($todoId);
		return new Tonic\Response(200);
	}
}


/**
 * @uri /todos/{todoId}/notes
 */
class TodoNotesCollection extends Framework\Resource {
	/**
	 * @method GET
	 * @provides application/json
	 */
	public function getList($todoId) {
		$repository = new Repositories\TodoRepository($this->dataBase);
		$notes = $repository->getNotesList($todoId);
		return new Tonic\Response(200, $notes);
	}
	 
	 /**
	  * @method POST
	  * @accepts application/json
	  */
	public function create($todoId) {
		$repository = new Repositories\TodoRepository($this->dataBase);
		
		$data = $this->request->data;
		$note = new Models\TodoNote;
		
		$note->UserId = $this->userId;
		$note->Note = $data->Note;
		$note->TodoId = $todoId;
		
		$repository->createNote($note);
		
		$response = new Tonic\Response(201);
		$response->Location = $this->baseUri . '/todos/' . $todoId . '/notes/' . $note->Id;
		
		return $response;
	}
	 
}

/**
 * @uri /todos/{todoId}/status
 */
class TodoStatusResource extends Framework\Resource {
	/**
	 * @method POST
	 * @accepts application/json
	 */
	public function updateStatus($todoId) {
		$repository = new Repositories\TodoRepository($this->dataBase,
			$this->restrictedUser,
			$this->userId
			);
		$data = $this->request->data;
		$repository->updateStatus($todoId, $data->Status);
		return new Tonic\Response(200);
	}
}

/**
 * @uri /todos/{todoId}/userId
 */
class TodoUserIdResource extends Framework\Resource {
	/**
	 * @method POST
	 * @accepts application/json
	 */
	public function updateUserId($todoId) {
		$repository = new Repositories\TodoRepository($this->dataBase,
			$this->restrictedUser,
			$this->userId
			);
		$data = $this->request->data;
		$repository->updateUserId($todoId, $data->UserId);
		return new Tonic\Response(200);
	}
}

?>