<?php

class ModelPropertyTypeError extends Exception {
	public function __construct($modelClass, $propName, $typeSpec) {
		$this->modelClass = $modelClass;
		$this->propName = $propName;
		$this->typeSpec = $typeSpec;

		$this->message =  "Invalid type specification '"
			. $this->typeSpec
			. "' for "
			. $this->modelClass
			. "::"
			. $this->propName;
	}
	
	public function getModelClass() {
		return $this->modelClass;
	}
	
	public function getPropertyName() {
		return $this->propName;
	}
	
	public function getTypeSpecification() {
		return $this->typeSpec;
	}
	
	private $modelClass;
	private $propName;
	private $typeSpec;
}

abstract class ModelBindError extends HandlerError {
	public function __construct($modelClass, $propertyName) {
		parent::__construct(HTTPStatus::BadRequest, NULL);
		$this->modelClass = $modelClass;
		$this->propName = $propertyName;
	}
	
	public function getModelClass() {
		return $this->modelClass;
	}
	
	public function getPropertyName() {
		return $this->propName;
	}
	
	private $modelClass;
	private $propName;
}


class ModelBindTypeError extends ModelBindError {
	public function __construct($modelClass, $propertyName, $propertyType, $value) {
		parent::__construct($modelClass, $propertyName);
		$this->message = "Could not convert '$value' to $propertyType for $modelClass::$propertyName";
	}
}

class ModelBindNullError extends ModelBindError {
	public function __construct($modelClass, $propertyName) {
		parent::__construct($modelClass, $propertyName);
		$this->message = "$modelClass::$propertyName cannot be NULL";
	}
}

class ModelBindNotFoundError extends ModelBindError {
	public function __construct($modelClass, $propertyName) {
		parent::__construct($modelClass, $propertyName);
		$this->message = "No value found for $modelClass::$propertyName";
	}
}


class ModelPropertyTypesHelper {
	public function __construct($className, array $typeMap) {
		if (!is_array($typeMap)) {
			return;
		}
		
		$this->modelClass = $className;
		$this->typeMap = $typeMap;
		$this->cache = array();
	}
	
	public function getPropertyInfo($propertyName) {
		if (!array_key_exists($propertyName, $this->typeMap)) {
			return NULL;
		}
		
		$typeSpec = $this->typeMap[$propertyName];
		
		if (array_key_exists($typeSpec, $this->cache)) {
			return $this->cache[$typeSpec];
		}
		
		if (substr($typeSpec, 0, 1) === "^") {
			$typeSpec = substr($typeSpec, 1);
			$passObject = true;
		} else {
			$passObject = false;
		}
		
		if (substr($typeSpec, -1) === "?") {
			$typeName = substr($typeSpec, 0, -1);
			$nullable = TRUE;
		} else {
			$typeName = $typeSpec;
			$nullable = FALSE;
		}
		
		try {
			$typeInfo = new ReflectionClass($typeName);
		} catch (Exception $e) {
			throw new ModelPropertyTypeError($this->modelClass, $propertyName, $typeSpec);
		}
		
		$ctor = $typeInfo->getConstructor();
		
		if (is_null($ctor)) {
			throw new ModelPropertyTypeError($this->modelClass, $propertyName, $typeSpec);
		}
		
		$ctorParams = $ctor->getParameters();
		
		if (count($ctorParams) != 1 || $ctorParams[0]->getClass()) {
			throw new ModelPropertyTypeError($this->modelClass, $propertyName, $typeSpec);
		}

		return $this->cache[$typeSpec] = array($typeInfo, $nullable, $passObject);
	}
	
	private $modelClass;
	private $typeMap;
	private $cache;
}


abstract class Model {
	protected function getPropertyTypes() {
    return array();
  }
  
  protected function validate() {
    return TRUE;
  }
  
  public function copyProperties($from) {
    $source = get_object_vars($from);
    
    foreach ($this as $property => $value) {
      if (array_key_exists($property, $source)) {
        $this->$property = $source[$property];
      }
    }
  }
  
  public function toXML(DOMDocument $document, $elementName) {
    return Model::toXMLElement($this, $document, $elementName);
  }
  
	public static function toXMLElement($model, DOMDocument $document, $elementName) {
    $classInfo = new ReflectionClass($model);
    $props = $classInfo->getProperties(/*ReflectionProperty::IS_PUBLIC*/);
    
    $element = $document->createElement($elementName);

    foreach ($props as $propInfo) {
      $propElement = $document->createElement($propInfo->getName(), $propInfo->getValue($model));
      $element->appendChild($propElement);
    }
    
    return $element;
	}
	
	public function bind($requestParams) {
		$classInfo = new ReflectionClass($this);

		$typesHelper = new ModelPropertyTypesHelper($classInfo->getName(), $this->getPropertyTypes());
		
		$props = $classInfo->getProperties();
		
		foreach ($props as $propInfo) {
			if (!$propInfo->isDefault() || !$propInfo->isPublic() || $propInfo->isStatic()) {
				continue;
			}
			
			$valueFound = FALSE;
			$value = NULL;
			
			if (array_key_exists($propInfo->name, $requestParams)) {
				$value = $requestParams[$propInfo->name];
				if ($value === "") {
					$value = NULL;
				}
				$valueFound = TRUE;
			} else {
				$value = $propInfo->getValue($this);
				
				if ($value !== NULL) {
					$valueFound = TRUE;
				}
			}
			
			if (!$valueFound) {
				throw new ModelBindNotFoundError($classInfo->name, $propInfo->name);
			}
			
			list ($typeInfo, $nullable, $passObject) = $typesHelper->getPropertyInfo($propInfo->name);
			
			if (is_null($typeInfo) || ($nullable && is_null($value))) {
				$propInfo->setValue($this, $value);
				continue;
			}
			
			if (!$nullable && is_null($value)) {
				throw new ModelBindNullError($classInfo->name, $propInfo->name);
			}
			
			$typeName = $typeInfo->getName();
			
			try {
				$object = new $typeName($value);
			} catch (TypeConversionError $e) {
				throw new ModelBindTypeError($classInfo->name, $propInfo->name, $typeName, $value);
			}
			
			$value = $passObject ? $object : $object->value;
			
			$propInfo->setValue($this, $value);
		}
	}
}

?>