<?php

class StatsDefinition extends Model {
	public $name;
	public $oid;
	public $color;
	public $graphType;
	public $graphTitle;
	public $verticalLabel;
	public $showVerticalLabel;
	public $verticalAxisType;
	public $autoScale;
	public $useDelta;
	public $multiplier;
	public $upper;
	public $lower;
	
	public function getPropertyTypes() {
		return array();
	}
}

class StatsDefinitionHandler extends Handler {
	private function formatModel($model) {
		$document = $this->getResponseDocument();
		$statsDef = $document->createElement("StatsDefinition");
		
		$this->addChild($statsDef, "name", $model->name);
		$this->addChild($statsDef, "oid", $model->oid);
		$this->addChild($statsDef, "color", $model->color);
		$this->addChild($statsDef, "graphType", $model->graphType);
		$this->addChild($statsDef, "graphTitle", $model->graphTitle);
		$this->addChild($statsDef, "verticalLabel", $model->verticalLabel);
		$this->addChild($statsDef, "showVerticalLabel", $model->showVerticalLabel);
		$this->addChild($statsDef, "verticalAxisType", $model->verticalAxisType);
		$this->addChild($statsDef, "autoScale", $model->autoScale);
		$this->addChild($statsDef, "useDelta", $model->useDelta);
		$this->addChild($statsDef, "multiplier", (float)$model->multiplier);
		$this->addChild($statsDef, "upper", $model->upper);
		$this->addChild($statsDef, "lower", $model->lower);
		
		return $statsDef;
	}
	
	private function queryStatsDefinition($id = null) {
		$querySQL = <<<EOQ
SELECT
	StatsDef.id,
	StatsDef.name,
	oid,
	color,
	graphType,
	title as graphTitle,
	verticalLabel,
	showVerticalLabel,
	verticalAxisType,
	autoScale,
	useDelta,
	multiplier,
	upper,
	lower
FROM
	StatsDef
LEFT JOIN 
	GraphTemplate
ON
	defaultGraphTemplateId = GraphTemplate.id
WHERE
	NOT deleted
EOQ;

		$db = $this->getDBConnection();
		
		if ($id !== null) {
			$querySQL .= "\tAND StatsDef.id = " . $id;
		}
		
		try {
			$rows = $this->selectObjects($querySQL);
		} catch (EmptyResultSetError $e) {
			$this->setStatus(HTTPStatus::NotFound);
			throw $e;
		}
		
		if ($id !== null) {
			return $rows[0];
		}
		
		return $rows;
	}
	
	public function getStatsDefinitionList() {
		$statsDefs = $this->queryStatsDefinition();
		
		$document = $this->getResponseDocument();
		$root = $this->createRoot();
		
		$container = $document->createElement("StatsDefinitions");
		$root->appendChild($container);
		
		foreach ($statsDefs as $model) {
			$element = $this->formatModel($model);
			$element->setAttribute("id", $model->id);
			$container->appendChild($element);
		}
	}
	
	public function getStatsDefinition(IntParam $id) {
		$statsDef = $this->queryStatsDefinition($id);
		
		$element = $this->formatModel($statsDef);
		$element->setAttribute("id", $id);
		$root = $this->createRoot();
		$root->appendChild($element);
	}
	
	private function saveGraphTemplate(StatsDefinition $model) {
	
		$graphTemplateId = NULL;
		
		if ($model->color != ""
			|| $model->graphType != ""
			|| $model->graphTitle != ""
			|| $model->verticalLabel != ""
			|| $model->showVerticalLabel != ""
			|| $model->verticalAxisType != ""
			|| $model->autoScale != ""
			|| $model->useDelta != ""
			|| $model->multiplier != ""
			|| $model->upper != ""
			|| $model->lower != "") {
			
			$preparedSQL = <<<EOQ
INSERT INTO GraphTemplate (
	title,
	color,
	graphType,
	verticalLabel,
	showVerticalLabel,
	verticalAxisType,
	autoScale,
	useDelta,
	multiplier,
	upper,
	lower
) VALUES (
	?,
	?,
	?,
	?,
	?,
	?,
	?,
	?,
	?,
	?,
	?
)
EOQ;
			$db = $this->getDBConnection();
			
			$statement = $db->prepare($preparedSQL);
			$statement->bind_param("sissisiiddd",
				$model->graphTitle,
				$model->color,
				$model->graphType,
				$model->verticalLabel,
				$model->showVerticalLabel,
				$model->verticalAxisType,
				$model->autoScale,
				$model->useDelta,
				$model->multiplier,
				$model->upper,
				$model->lower);
			$statement->execute();
			
			$graphTemplateId = $statement->insert_id;
			
			if ($graphTemplateId == 0) {
				throw new HandlerError(HTTPStatus::BadRequest, $statement->error);
			}
		}

		return $graphTemplateId;
	}
	
	public function postStatsDefinition(StatsDefinition $model) {
		$db = $this->getDBConnection();
		
		// Start a transaction
		
		$db->autocommit(FALSE);
		
		// Insert the graph template
		
		$graphTemplateId = $this->saveGraphTemplate($model);
		
		// Insert the stats definition
		
		$preparedSQL = <<<EOQ
INSERT INTO StatsDef (
	name,
	oid,
	defaultGraphTemplateId
) VALUES (
	?,
	?,
	?
)	
EOQ;

		$statement = $db->prepare($preparedSQL);
		$statement->bind_param("ssi", $model->name, $model->oid, $graphTemplateId);
		$statement->execute();
		
                if ($statement->errno === 1062) {
                        throw new HandlerError(HTTPStatus::Conflict, $statement->error);
                }

                if ($statement->errno !== 0) {
                        throw new HandlerError(HTTPStatus::InternalError, $statement->error);
                }

		// Commit the transaction 
		$db->commit();

		$statsDefId = $statement->insert_id;

		$this->setStatus(HTTPStatus::Created);
		
		Header("Location: " . $_SERVER["REQUEST_URI"] . "/" . $statsDefId);
		
		$element = $this->formatModel($model);
		$element->setAttribute("id", $statsDefId);
		
		$root = $this->createRoot();
		$root->appendChild($element);
	}
	
	private function deleteGraphTemplate($statsDefId) {
		$deleteTemplateSQL = <<<EOQ
DELETE 
	GraphTemplate
FROM 
	StatsDef 
INNER JOIN 
	GraphTemplate 
ON 
	StatsDef.defaultGraphTemplateId = GraphTemplate.id 
WHERE
	StatsDef.id = $statsDefId;
EOQ;
		$this->query($deleteTemplateSQL);
	}
	
	public function putStatsDefinition(IntParam $id, StatsDefinition $model) {
		
		$db = $this->getDBConnection();
	
		// Begin a transaction
		
		$db->autocommit(FALSE);
		
		// Delete the graph template record if present
		
		$this->deleteGraphTemplate($id);
		
		// Create new graph template record if necessary
		
		$graphTemplateId = $this->saveGraphTemplate($model);
		
		// Update the stats definition record
		
		$preparedSQL = <<<EOQ
UPDATE
	StatsDef
SET
	name = ?,
	oid = ?,
	defaultGraphTemplateId = ?
WHERE
	id = $id
EOQ;
		$statement = $db->prepare($preparedSQL);
		$statement->bind_param("ssi", $model->name, $model->oid, $graphTemplateId);
		$statement->execute();
		
		// Commit the transaction
		
		$db->commit();
		
		if ($statement->affected_rows == 0) {
			if ($statement->errno) {
				throw new HandlerError(HTTPStatus::BadRequest, $statement->error);
			}
			
			$this->setStatus(HTTPStatus::NotFound);
			return;
		}
		
		$root = $this->createRoot();
		
		$element = $this->formatModel($model);
		$element->setAttribute("id", $id);
		$root->appendChild($element);
	}
	
	public function deleteStatsDefinition(IntParam $id) {
		$db = $this->getDBConnection();
		$db->autocommit(FALSE);
		
		$this->deleteGraphTemplate($id);
		
		$deleteStatsDefSQL = "DELETE FROM StatsDef WHERE id = " . $id;
		$db->query($deleteStatsDefSQL);
		
		$rowsDeleted = $db->affected_rows;
		
		$db->commit();
		
		if ($rowsDeleted == 0) {
			$this->setStatus(HTTPStatus::NotFound);
		}
	}
}

?>