<?php

//--------------------------------------------------------------
// Low level micro-framework for developing RESTful XML web
// services for WNMS. To be documented and commented.
// Functionality of the framework depends on a URL rewrite rule
// in the Apache config file for the WNMS website that passes
// database connection parameters to this PHP file through
// Apache's environment variable.
//
// Copyright (C) 2011 LigoWave, all rights reserved.
//--------------------------------------------------------------


class TypeConversionError extends Exception {
	public function __construct($typeName, $value) {
		
		$this->typeName = $typeName;
		$this->value = $value;
		
		$this->message = ($value === NULL ? "NULL" : "'$value'") 
			. " is not a valid value for type $typeName";
	}
	
	public function getTypeName() {
		return $this->typeName;
	}
	
	public function getValue() {
		return $this->value;
	}
	
	private $typeName;
	private $value;
}

class DatabaseError extends Exception {
	public function __construct($errno, $error) {
	
		$this->errno = $errno;
		$this->error = $error;
		
		$this->message = "Database error: " . $error;
	
	}
	
	public function getErrno() {
		return $this->errno;
	}
	
	public function getError() {
		return $this->error;
	}
	
	private $errno;
	private $error;
}

class DatabaseConnectError extends DatabaseError {
	public function __construct($connectErrno, $connectError) {
		parent::__construct($connectErrno, $connectError);
	}
}

class EmptyResultSetError extends Exception {
	public function __construct($sql) {
		$this->sql = $sql;
		$this->message = "Database query returned an empty result set";
	}
	
	public function getSQL() {
		return $this->sql;
	}
}

class HTTPStatus {
	const OK = "200 OK";
	const Created = "201 Created";
	const BadRequest = "400 Bad Request";
	const Unauthorized = "401 Unauthorized";
	const NotFound = "404 Not Found";
	const Conflict = "409 Conflict";
	const UnsupportedMediaType = "415 Unsupported Media Type";
	const InternalServerError = "500 Internal Server Error";
	const ServiceUnavailable = "503 Service Unavailable";
}

class HandlerError extends Exception {
	public function __construct($httpStatus, $message) {
		$this->httpStatus = $httpStatus;
		$this->message = $message;
	}
	
	public function getHTTPStatus() {
		return $this->httpStatus;
	}
}

class AuthorizationError extends HandlerError {
	public function __construct() {
		parent::__construct(HTTPStatus::Unauthorized,
			"Authorization required");
	}
}

//
// Since PHP type hinting does not work with built-in primitive
// types, handler classes can use the IntParam class to ensure
// type safety of integer parameters passed to their public
// service methods
//

abstract class Parameter {
	public $value;

	abstract public function __construct($value);
	
	public function __toString() {
		return (string) $this->value;
	}
}

class IntParam extends Parameter {
	public function __construct($value) {
		$this->value = intval($value);
		if ((string)$this->value != $value) {
			throw new TypeConversionError(__CLASS__, $value);
		}
	}
}

class DateParam extends Parameter {
	public $year;
	public $month;
	public $day;

	public function __construct($value) {
		$matches = array();
		
		if (preg_match("/^(\d{4})-(\d{1,2})-(\d{1,2})$/", $value, $matches) == 0) {
			throw new TypeConversionError(__CLASS__, $value);
		}
		
		$this->year = (int) $matches[1];
		$this->month = (int) $matches[2];
		$this->day = (int) $matches[3];
		
		if (!checkdate($this->month, $this->day, $this->year)) {
			throw new TypeConversionError(__CLASS__, $value);
		}

		$this->value = sprintf("%d-%02d-%02d",
			$this->year,
			$this->month,
			$this->day);
	}
}

class RouteInfo {
  public $handlerName;
  public $actionName;
  public $resourceId;
  public $isCollection;
  public $pathRemainder;
}

//
// The Handler class serves both as the implementation of the
// framework (in its static methods) and the base class for
// service handler classes, providing a number of useful utility
// functions for manipulating the response XML document and
// querying the database.
//
// This class is currently directly tied to WNMS because it
// opens a connection to the WNMS database and performs 
// WNMS-specific request authorization. This can be worked
// around by making the authorize() method an instance method
// (as opposed to static), then deriving a class named
// WNMSHandlerBase that overrides init() and authorize() and
// implements the DB connection and authorization logic there,
// and then have all the actual handler classes extend
// WNMSHandlerBase instead of Handler. However, there is
// currently no need for such abstraction and it is best to 
// keep things simple.
//

class Handler {

	public static function run() {
		ob_start();

		$document = new DOMDocument();
		$document->formatOutput = true;
		
		spl_autoload_register(array("self", "loadClass"));
		
		$exceptionCaught = NULL;

		try {
			self::dispatchRequest($document);
		} catch (HandlerError $e) {
			self::setStatus($e->getHTTPStatus());
			$exceptionCaught = $e;
		} catch (DatabaseConnectError $e) {
			self::setStatus(HTTPStatus::ServiceUnavailable);
			$exceptionCaught = $e;
		} catch (DatabaseError $e) {
			self::setStatus(HTTPStatus::InternalServerError);
			$exceptionCaught = $e;
		} catch (Exception $e) {
			$exceptionCaught = $e;
		}
		
		if ($exceptionCaught) {
			$document = self::formatError($exceptionCaught);
		}
		
		$output = ob_get_contents();
		
		ob_end_clean();
		
		if (strlen($output)) {
			$comment = $document->createComment($output);
			$comment = $document->appendChild($comment);
		}

		if (!is_null(self::$httpStatus)) {
			// Work around a design flaw in Flex 3.5/Flash
			// that prevents ActionScript code from ever
			// handling any 4xx responses
			if (self::$httpStatus[0] === "4" && self::getRequestHeader("X-HTTP-4xx-Status-Workaround")) {
				header("X-HTTP-Status: " . self::$httpStatus);
				self::$httpStatus = HTTPStatus::OK;
			}
		
			header("HTTP/1.1 " . self::$httpStatus);
		}
		
		if ($document->hasChildNodes()) {
			header("Content-Type: application/xml;charset=UTF-8");
			echo $document->saveXML();
		}
	}
	
	private static function loadClass($className) {
		$sourceCode = dirname(__FILE__) . DIRECTORY_SEPARATOR . $className . ".php";
		
		if (file_exists($sourceCode)) {
			include $sourceCode;
		} else {
			throw new HandlerError(HTTPStatus::NotFound, "Class $className not found");
		}
	}
	
	private static function formatError(Exception $e) {
		$message = $e->getMessage();
		$message = trim($message);
		
		$document = new DOMDocument();
		$document->formatOutput = true;
		
		$root = $document->createElement("Result");
		$root = $document->appendChild($root);
		
		$root->setAttribute("success", "false");
		
		$error = $document->createElement("Error");
		$error = $root->appendChild($error);
		
		$element = $document->createElement("Type", get_class($e));
		$element = $error->appendChild($element);
		
		$element = $document->createElement("Message", $message);
		$element = $error->appendChild($element);
		
		$location = $e->getFile() . ":" . $e->getLine();
		
		$element = $document->createElement("Location", $location);
		$element = $error->appendChild($element);
		
		$traceNode = $document->createCDATASection($e->getTraceAsString());
		
		$element = $document->createElement("Trace");
		$element = $error->appendChild($element);
		
		$element->appendChild($traceNode);
		return $document;
	}
	
	private static function parseRoute($path) {
		$route = new RouteInfo();
		
		$matches = array();
		
		if  (preg_match("|^/(\w+)s/(\d+)/?$|", $path, $matches)) {
			$route->handlerName = $matches[1];
			$route->resourceId = $matches[2];
			$route->actionName = "";
			$route->isCollection = TRUE;
			$route->pathRemainder = NULL;
		} else if (preg_match("|^/(\w+)s/(\d+)/(\w+)s?(/.*)$|U", $path, $matches)) {
      $route->handlerName = $matches[1];
      $route->resourceId = $matches[2];
      $route->actionName = $matches[3];
      $route->isCollection = TRUE;
      $route->pathRemainder = $matches[4];
		} else if (preg_match("|^/(\w+)s/(\d+)/(\w+)$|U", $path, $matches)) {
			$route->handlerName = $matches[1];
			$route->resourceId = $matches[2];
			$route->actionName = $matches[3];
			$route->isCollection = TRUE;
			$route->pathRemainder = NULL;
		} else if (preg_match("|^/(\w+)s/?$|U", $path, $matches)) {
			$route->handlerName = $matches[1];
			$route->resourceId = "";
			$route->actionName = "";
			$route->isCollection = TRUE;
			$route->pathRemainder = NULL;
		} else if (preg_match("|^/(\w+)s?/(\w+)/?$|U", $path, $matches)) {
			$route->handlerName = $matches[1];
			$route->resourceId = "";
			$route->actionName = $matches[2];
			$route->isCollection = preg_match("|^/\w+s/|", $path);
			$route->pathRemainder = NULL;
		} else {
			throw new HandlerError(HTTPStatus::NotFound, "The requested URI does not match any of the built-in routes");
		}
		
		$route->handlerName = ucfirst($route->handlerName);
		$route->actionName = ucfirst($route->actionName);
		
		return $route;
	}
	
	private static function buildParameters($methodInfo) {
		$methodParams = $methodInfo->getParameters();
		$paramValues = array();
		
		$requestParams = array_merge($_GET, $_POST);
		
		foreach ($methodParams as $parameter) {
			$paramClass = $parameter->getClass();
			$paramName = $parameter->name;
			if ($paramClass && $paramClass->name == "DOMDocument") {
				$xmlContentTypes = array("text/xml", "application/xml");
				$xml = self::getRequestContentAs($xmlContentTypes);
				if (!$xml) {
					throw new HandlerError(HTTPStatus::UnsupportedMediaType,
						"Content type application/xml expected");
				}
				$paramValue = new DOMDocument();
				if (DOMDocument::loadXML($xml) === FALSE) {
					throw new HanderError(HTTPStatus::BadRequest,
						"Could not parse request content as XML");
				}
			} else if ($paramClass && $paramClass->isSubclassOf("Model")) {
				$paramValue = new $paramClass->name();
				$paramValue->bind($requestParams);
			} else if (isset($requestParams[$paramName])) {
				$paramValue = $requestParams[$paramName];
			} else if ($parameter->isDefaultValueAvailable()) {
				$paramValue = $parameter->getDefaultValue();
			} else {
				throw new HandlerError(HTTPStatus::BadRequest, "Argument $paramName not found");
			}
			
			if ($paramClass && $paramClass->isSubclassOf("Parameter")) {
				$className = $paramClass->name;
				
				try {
					$paramValue = new $className($paramValue);
				} catch (TypeConversionError $e) {
					throw new HandlerError(HTTPStatus::BadRequest,
						"Parameter '$paramName': " . $e->getMessage());
				}
			}
			
			$paramValues[] = $paramValue;
		}
		
		return $paramValues;
	}
	
	private static $requestHeaders;
	private static $requestContentType;
	private static $requestContent;
	
	private static function fetchRequestHeaders() {
		if (!is_array(self::$requestHeaders)) {
			self::$requestHeaders = apache_request_headers();
			self::$requestHeaders = array_change_key_case(self::$requestHeaders, CASE_LOWER);
		}
	}
	
	private static function getRequestHeader($headerName) {
		self::fetchRequestHeaders();
		$name = strtolower($headerName);
		return array_key_exists($name, self::$requestHeaders)
			? self::$requestHeaders[$name]
			: null;
	}
	
	private static function getRequestMethod() {
		self::fetchRequestHeaders();
		
		if (array_key_exists("x-http-method-override", self::$requestHeaders)) {
			$override = strtoupper(self::$requestHeaders["x-http-method-override"]);
			return $override;
		}
		
		return $_SERVER["REQUEST_METHOD"];
	}
	
	private static function getRequestContentType() {
		if (self::$requestContentType !== NULL) {
			return self::$requestContentType;
		}
		
		self::fetchRequestHeaders();

		if (!array_key_exists("content-type", self::$requestHeaders)) {
			return NULL;
		}
		
		list(self::$requestContentType) = explode(";", self::$requestHeaders["content-type"], 1);
		
		return self::$requestContentType;
	}
	
	protected static function getRequestContent($cache = TRUE) {
		if (self::$requestContent !== NULL) {
			return self::$requestContent;
		}
		
		self::fetchRequestHeaders();
		
		$encoding = NULL;
		
		if (array_key_exists("content-transfer-encoding", self::$requestHeaders)) {
			$supportedEncodings = array("base64", "quoted-printable", "8bit", "7bit", "binary");

			$encoding = self::$requestHeaders["content-transfer-encoding"];
			$encoding = strtolower($encoding);
			
			if (!in_array($encoding, $supportedEncodings)) {
				return NULL;
			}
		}

		$content = file_get_contents("php://input");
		
		if ($encoding == "base64") {
			$content = base64_decode($content);
			if ($content === FALSE) {
				return NULL;
			}
		} else if ($encoding == "quoted-printable") {
			$content = quoted_printable_decode($content);
		}
		
		if ($cache) {
			self::$requestContent = $content;
		}
		
		return $content;
	}
	
	private static function getRequestContentAs($contentType, $cache = TRUE) {
		self::fetchRequestHeaders();
		
		$requestContentType = self::getRequestContentType();
		
		if (is_array($contentType)) {
			$contentTypeMatches = in_array($requestContentType, $contentType);
		} else {
			$contentTypeMatches = ($requestContentType == $contentType);
		}
		
		if (!$contentTypeMatches) {
			return NULL;
		}
		
		return self::getRequestContent($cache);
	}

	private static function authorize(mysqli $dbConnection) {
		session_start();
		
		$userId = 0;
		
		if (isset($_SESSION["uguid"])) {
			$result = $dbConnection->query("SELECT id FROM Users WHERE guid='" . $dbConnection->escape_string($_SESSION["uguid"]) . "'");
			if ($result) {
				$row = $result->fetch_row();
				$result->free();
				$userId = (int) $row[0];
			}
		}
		
		return ($userId != 0);
	}
	
	private static function databaseConnect() {
		
		$connectionString = apache_getenv("DBPARAMS");
		
		if ($connectionString === FALSE) {
			throw new Exception("Database connection parameters unavailable");
		}
		
		$connectionParams = explode(":", $connectionString);
		
		if (count($connectionParams) != 4) {
			throw new Exception("Database connection parameters format error");
		}
		
		list($host, $user, $pass, $dbname) = $connectionParams;
		
		$dbConnection = mysqli_init();

		$connectionTimeout = ini_get("max_execution_time");
		
		if ($connectionTimeout == 0 || $connectionTimeout > 10) {
			$connectionTimeout = 10;
		}
		
		$dbConnection->options(MYSQLI_OPT_CONNECT_TIMEOUT, $connectionTimeout);

		if (!$dbConnection->real_connect($host, $user, $pass, $dbname)) {
			throw new DatabaseConnectError($dbConnection->connect_errno, $dbConnection->connect_error);
		}
		
		$dbConnection->set_charset("utf8");

		return $dbConnection;
	}

	private static function dispatchRequest($responseDocument) {
		$reqMethod = self::getRequestMethod();
		$reqMethod = strtolower($reqMethod);

    $classInfo = NULL;
    $handler = NULL;
    
    $path = $_SERVER["PATH_INFO"];
    
    $parseRoute = TRUE;
    $lastRoute = NULL;
    
    while ($parseRoute) {
      $route = self::parseRoute($path);
      
      if ($lastRoute !== NULL) {
        $actionName = preg_replace("/s$/", "", $lastRoute->actionName);
        $methodName = lcfirst($actionName) . "Handler";
        
        if ($classInfo->hasMethod($methodName)) {
          $methodInfo = new ReflectionMethod($classInfo->name, $methodName);

          if (!$methodInfo->isStatic()) {
            throw new HandlerError(HTTPStatus::InternalServerError,
              "{$classInfo->name}::{$methodName}() must be static");
          }
         

          $parameters = $methodInfo->getParameters();
          $paramName = lcfirst($lastRoute->handlerName) . "Id";

          if (count($parameters) != 1 
            || $parameters[0]->name != $paramName
            || $parameters[0]->getClass() === NULL
            || $parameters[0]->getClass()->name !== "IntParam"
          ) {
            throw new HandlerError(HTTPStatus::InternalServerError,
              "{$classInfo->name}::{$methodName}() must accept a single integer argument {$paramName}");
          }
          
          $handler = $methodInfo->invoke(null, new IntParam($lastRoute->resourceId));
          
          $classInfo = new ReflectionClass($handler);
          $className = $classInfo->name;
          
          if (!$classInfo->isSubclassOf("Handler")) {
            throw new HandlerError(HTTPStatus::InternalServerError,
              "{$className} is not a subclass of Handler");
          }
        } else {
          throw new HandlerError(HTTPStatus::InternalServerError,
            "{$classInfo->name}::{$methodName}() not found");
        }
      }
      
      if ($handler === NULL) {
        $className = $route->handlerName . "Handler";
        $classInfo = new ReflectionClass($className);
        
        if (!$classInfo->isSubclassOf("Handler")) {
          throw new HandlerError(HTTPStatus::NotFound, "$className is not a subclass of Handler");
        }
      }
      
      if ($route->resourceId !== ""
        && $route->actionName === ""
        && $route->isCollection 
        && $reqMethod === "post") {
        $reqMethod = "update";
      }

      $methodName = $reqMethod . (
        $route->actionName == "" ? $route->handlerName : $route->actionName
      );

      if ($route->resourceId == "" && $route->isCollection) {
        if ($reqMethod == "get") {
          $methodName = $reqMethod . (
            $route->actionName == "" 
              ? $route->handlerName . "List" 
              : $route->actionName
          );
        } else if ($reqMethod != "post") {
          throw new HandlerError(HTTPStatus::NotFound, "Only GET and POST requests to collection URIs are allowed");
        }
      }
      
      $needAuth = TRUE;

      $noauthMethodName = "noauth" . ucfirst($methodName);
		
      try {
        $methodInfo = new ReflectionMethod($className, $noauthMethodName);
        $needAuth = FALSE;
      } catch (Exception $e) {
        $methodInfo = NULL;
      }
      
      if ($methodInfo === NULL) {
        try {
          $methodInfo = new ReflectionMethod($className, $methodName);
        } catch (Exception $e) {
        
          if ($route->actionName !== "" && $route->resourceId !== "") {
            $actionName = preg_replace("/s$/", "", $route->actionName);
            $path = "/{$actionName}s{$route->pathRemainder}";
            $lastRoute = $route;
            continue;
          }
        
          throw new HandlerError(HTTPStatus::NotFound,
            "Neither $className::$noauthMethodName() nor $className::$methodName() found");
        }
      }
      
      $parseRoute = FALSE;
    }
		
		if (count($_POST) === 0 && self::getRequestContentType() === "application/x-www-form-urlencoded") {
			parse_str(self::getRequestContent(), $_POST);
		}
		
		if ($route->resourceId != "") {
			$methodParams = $methodInfo->getParameters();
			$hasIdParam = FALSE;
			
			foreach ($methodParams as $parameter) {
				if ($parameter->name == "id") {
					$hasIdParam = TRUE;
					break;
				}
			}
			
			if (!$hasIdParam) {
				throw new HandlerError(HTTPStatus::NotFound,
					"$className::{$methodInfo->name} does not accept an 'id' parameter");
			}
			
			$_POST["id"] = $route->resourceId;
		}
		
		$args = self::buildParameters($methodInfo);
		
		$dbConnection = self::databaseConnect();
		
		if ($needAuth && !self::authorize($dbConnection)) {
			throw new AuthorizationError();
		}
		
		if ($handler === NULL) {
      $handler = new $className();
    }
    
    $handler->doc = $responseDocument;
    $handler->db = $dbConnection;

    $exceptionCaught = NULL;
    $returnValue = NULL;

    $handler->init();
    
    try {
      $returnValue = $methodInfo->invokeArgs($handler, $args);
    } catch (Exception $e) {
      $exceptionCaught = $e;
    }
    
    $handler->cleanup();
    
    if ($exceptionCaught === NULL && $returnValue !== NULL) {
      $classInfo = new ReflectionClass($returnValue);
      if ($classInfo->isSubclassOf("Handler")) {
        $handler = $returnValue;
        $chain = TRUE;
      }
    }

		$dbConnection->close();
		
		if ($exceptionCaught) {
			throw $exceptionCaught;
		}
	}

	private $doc;
	private $db;
	
	private static $httpStatus;
	
	protected static function setStatus($httpStatus) {
		self::$httpStatus = $httpStatus;
	}
	
	protected function init() {
	}
	
	protected function cleanup() {
	}
	
	protected final function getDBConnection() {
		return $this->db;
	}
	
	protected final function getResponseDocument() {
		return $this->doc;
	}
	
	protected final function query($sqlCode) {
		$result = $this->db->query($sqlCode);
		
		if ($result === FALSE) {
			throw new DatabaseError($this->db->errno, $this->db->error);
		}

		return $result;
	}
	
	protected final function selectSingle($sqlCode) {
		$result = $this->query($sqlCode);
		$rowArr = $result->fetch_row();
		$result->free();
		
		if ($rowArr === NULL) {
			throw new EmptyResultSetError($sqlCode);
		}
		
		return $rowArr[0];
	}
	
	protected final function selectObject($sqlCode) {
		$result = $this->query($sqlCode);
		$object = $result->fetch_object();
		$result->free();
		
		if ($object === NULL) {
			throw new EmptyResultSetError($sqlCode);
		}
		
		return $object;
	}
	
	protected final function selectObjects($sqlCode) {
		$objects = array();
		$result = $this->query($sqlCode);
		while ($object = $result->fetch_object()) {
			$objects[] = $object;
		}
		$result->free();
		
		if (count($objects) == 0) {
			throw new EmptyResultSetError($sqlCode);
		}
		
		return $objects;
	}

	protected final function escapeString($s) {
		return $this->db->escape_string($s);
	}
	
	protected final function quoteString($s) {
		if ($s === NULL) {
			return "NULL";
		}
		
		return "'" . $this->escapeString($s) . "'";
	}
	
	protected final function truncateDocument() {
		while ($this->doc->firstChild) {
			$this->doc->removeChild($this->doc->firstChild);
		}
	}

	protected final function createRoot() {
		$this->truncateDocument();
		
		$root = $this->doc->createElement("Result");
		$root = $this->doc->appendChild($root);
		
		$root->setAttribute("success", "true");
		
		return $root;
	}
	
	protected final function addChild($parent, $name, $value = NULL) {
		$child = ($value === NULL)
			? $this->doc->createElement($name)
			: $this->doc->createElement($name, $value);
		$child = $parent->appendChild($child);
		
		return $child;
	}
	
	protected final function addText($element, $text) {
		$textNode = $this->doc->createTextNode($text);
		$textNode = $element->appendChild($textNode);
		
		return $textNode;
	}
}

//
// Run Forrest, run!
//

Handler::run();
?>